import datetime
from typing import Iterable, Any, List
from abc import abstractmethod

import django_filters
import xlwt

from .fields import FieldFileAbsoluteURL

META_REPORT_KEY = "use_for_report"
REPORT_FIELDS_KEY = "flex_report_search_fields"
REPORT_EXCULDE_KEY = "flex_report_search_exclude"
REPORT_COLUMNS_EXCULDE_KEY = "flex_report_columns_exclude"
REPORT_CUSTOM_FIELDS_KEY = "flex_report_custom_fields"

REPORT_CELL_STYLE_MAP = (
    (datetime.datetime, xlwt.easyxf(num_format_str="YYYY/MM/DD HH:MM")),
    (datetime.date, xlwt.easyxf(num_format_str="DD/MM/YYYY")),
    (datetime.time, xlwt.easyxf(num_format_str="HH:MM")),
    (bool, xlwt.easyxf(num_format_str="BOOLEAN")),
    (
        FieldFileAbsoluteURL,
        lambda v: xlwt.Formula(f'HYPERLINK("{v}","{v}")') if v else "",
    ),
)

FILTERSET_DATE_FILTERS = [
    django_filters.DateFilter,
    django_filters.TimeFilter,
    django_filters.DateTimeFilter,
]


class ReportModel:
    models = []

    @classmethod
    def register(cls, *models):
        cls.models.extend(models)
        return models[0]


class DynamicSubField:
    slug = None
    verbose_name = None

    def __init__(self, verbose_name=None):
        self.verbose_name = verbose_name or self.verbose_name

    def get_verbose_name(self):
        return str(self.verbose_name)

    @abstractmethod
    def get_value(cls, *args, **kwargs):
        pass


class BaseDynamicField:
    field_slug = None
    verbose_name = None
    model = None
    fields = {}

    @classmethod
    def get_by_slug(cls, slug):
        try:
            return cls.fields[slug]
        except KeyError as e:
            raise NotImplementedError(f"Field with slug {slug} not found") from e

    @classmethod
    def register(cls, field):
        assert issubclass(field, cls)
        cls.fields.update({field.field_slug: field})
        return field

    @classmethod
    @abstractmethod
    def unpack_field(cls, *args, **kwargs) -> List[DynamicSubField]:
        pass


class BaseExportFormat:
    formats = {}
    format_slug = None
    format_name = None
    format_ext = None
    export_headers = {}
    export_qs = []
    export_kwargs = {}
    export_filename = None

    def __init__(self, user=None, request=None):
        self.user = user
        self.request = request

    def get_export_headers(self):
        return self.export_headers

    def get_export_kwargs(self):
        return self.export_kwargs

    def get_export_filename(self):
        return f"{self.export_filename or self.format_slug}{self.format_ext}"

    def get_export_qs(self):
        return self.export_qs

    def check_auth(self):
        return True

    def __str__(self):
        return self.format_name

    @classmethod
    @property
    @abstractmethod
    def format_slug(cls):
        raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def format_name(cls):
        raise NotImplementedError

    @classmethod
    def register(cls, format_):
        assert issubclass(format_, BaseExportFormat)
        cls.formats.update({format_.format_slug: format_})
        return format_

    @classmethod
    def register_formats(cls, formats: dict):
        cls.formats.update(formats)

    @abstractmethod
    def handle(cls):
        raise NotImplementedError

    @abstractmethod
    def handle_response(self, *args, **kwargs):
        raise NotImplementedError


class FieldTypes:
    field = "field"
    property = "property"
    custom = "custom"
    dynamic = "dynamic"


class BaseDynamicColumn:
    columns = {}



# Generated by Django 4.2 on 2023-11-22 03:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flex_report", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="column",
            name="title",
            field=models.CharField(db_index=True, max_length=150, verbose_name="Title"),
        ),
    ]



from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin

from django import forms
from django.contrib import admin
from django.utils.html import format_html

from .models import Column, TableButton, TablePage, Template, TableButtonColor
from .utils import get_table_page_choices, get_table_page_optional_choices


@admin.register(TableButtonColor)
class TableButtonColorAdmin(admin.ModelAdmin):
    list_display = ["title", "color"]
    search_fields = ["title", "color"]


@admin.register(Template)
class TemplateAdmin(admin.ModelAdmin):
    list_display = [
        "title",
        "creator",
        "model",
        "created_date",
        "modified_date",
        "status",
        "is_page_default",
        "columns_count",
    ]
    raw_id_fields = ["creator"]
    search_fields = ["title", "creator", "model"]
    list_filter = ["is_page_default", "status"]


@admin.register(TablePage)
class TablePageAdmin(admin.ModelAdmin):
    readonly_fields = ("url",)
    list_display = ("title", "url")
    search_fields = readonly_fields + list_display

    @admin.display(description="URL Name")
    def url(self, obj):
        return format_html(f"<a href='{obj.url}'>{obj.url_name}</a>")

    def get_form(self, request, obj=None, **kwargs):
        kwargs["widgets"] = {"url_name": forms.Select(choices=get_table_page_choices())}
        return super().get_form(request, obj, **kwargs)

    def get_readonly_fields(self, _, obj=None):
        return self.readonly_fields if obj else ()


@admin.register(Column)
class ColumnAdmin(admin.ModelAdmin):
    search_fields = ["title"]
    list_display = ["title", "model"]


@admin.register(TableButton)
class TableButtonAdmin(admin.ModelAdmin, DynamicArrayMixin):
    list_readonly_fields = ["url", "title"]
    list_display = ["title", "display_name", "icon", "url", "color"]
    search_fields = ["url", "color"]
    list_editable = list(set(list_display) - set(list_readonly_fields))

    @admin.display(description="URL Name")
    def url(self, obj):
        return format_html(f"<a href='{obj.url}'>{obj.url_name}</a>")

    def get_form(self, request, obj=None, **kwargs):
        kwargs["widgets"] = {
            "url_name": forms.Select(choices=get_table_page_optional_choices())
        }
        return super().get_form(request, obj, **kwargs)

    def get_readonly_fields(self, _, obj=None):
        return self.readonly_fields if obj else ()



import ast
import mimetypes
from operator import call, ne, methodcaller, attrgetter
from functools import partial
from logging import getLogger

from django.core.exceptions import PermissionDenied
from django.db.models import Subquery
from django.core.paginator import EmptyPage, Paginator
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.shortcuts import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import View

from flex_report import export_format, BaseExportFormat

from .app_settings import app_settings
from .templatetags.flex_report_filters import get_column_verbose_name
from .filterset import (
    generate_filterset_from_model,
    generate_quicksearch_filterset_from_model,
)
from .models import Template
from .utils import (
    generate_filterset_form,
    get_template_columns,
    get_choice_field_choices,
    string_to_q,
    FieldTypes,
)

logger = getLogger(__file__)


class PaginationMixin(View):
    pages = [25, 75, 100, 200]
    default_page = pages[0]
    pagination = None
    page_keyword = "page"
    per_page_ketyword = "per_page"

    def get_page(self):
        page = self.request.GET.get(self.page_keyword, 1)
        per_page = (
            p
            if (p := self.request.GET.get(self.per_page_ketyword, self.default_page))
            and p in map(str, self.pages)
            else self.default_page
        )
        try:
            paginator = Paginator(self.get_paginate_qs(), per_page)
            page_obj = paginator.page(page)
        except EmptyPage:
            page_obj = paginator.page(1)
        return page_obj

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        page = self.get_page()
        context["pagination"] = self.pagination = {
            "pages": self.pages,
            "qs": page,
            "paginator": page.paginator,
            "page_keyword": self.page_keyword,
            "per_page_keyword": self.per_page_ketyword,
        }
        return context

    def get_paginate_qs(self):
        return []


class TemplateObjectMixin(View):
    template_object = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        try:
            self.template_object = self.get_template()
        except PermissionDenied:
            return HttpResponseForbidden()

    def dispatch(self, *args, **kwargs):
        from .models import Template

        handler = None
        match self.template_object and self.template_object.status:
            case Template.Status.complete:
                handler = self.template_ready()
            case Template.Status.pending:
                handler = self.template_not_ready()
        return handler or super().dispatch(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        return {
            "realtime_quicksearch": app_settings.REALTIME_QUICKSEARCH,
            "has_export": self.template_object.has_export,
        }

    def get_template(self):
        return self.get_object()

    def template_ready(self):
        pass

    def template_not_ready(self):
        pass


class QuerySetExportMixin(View):
    export_qs = []
    export_headers = {}
    export_columns = []
    export_kwargs = {}
    export_filename = None

    def get_export_filename(self):
        return self.export_filename

    def get_export_columns(self):
        return self.export_columns

    def get_export_headers(self):
        return self.export_headers

    def get_export_qs(self):
        return self.export_qs

    def get_export_kwargs(self):
        return self.export_kwargs

    def get_handle_qs(self):
        return {
            "export_qs": self.get_export_qs(),
            "export_headers": self.get_export_headers(),
            "export_columns": self.get_export_columns(),
            "export_kwargs": self.get_export_kwargs(),
        }

    def test_func(self):
        if not hasattr((exporter := self.get_exporter()), "test_func"):
            return

        if exporter.test_func():
            return

        raise HttpResponseForbidden(content="403 Forbidden")

    def dispatch(self, *args, **kwargs):
        if (
            not (format_ := self.request.GET.get("format", "").lower())
            or format_ not in export_format.formats.keys()
        ):
            return HttpResponseBadRequest()

        self.export_format = format_
        self.test_func()

        return super().dispatch(*args, **kwargs)

    def get_exporter(self) -> BaseExportFormat:
        try:
            format_ = export_format.formats[self.export_format]
            if any(self.get_handle_qs().values()):
                return type("DynamicExporter", (format_,), self.get_handle_qs())(
                    request=self.request, user=self.request.user
                )
            return format_(request=self.request, user=self.request.user)
        except KeyError as e:
            raise NotImplementedError(
                f"The wanted format '{self.export_format}' isn't handled."
            ) from e

    def get(self, *args, **kwargs):
        format_ = self.get_exporter()
        filename = str(format_.get_export_filename())

        response = HttpResponse(
            content_type=mimetypes.types_map.get(
                f".{format_.format_ext}",
                "application/octet-stream",
            ),
            headers={"Content-Disposition": f'attachment; filename="{filename}"'},
        )
        response = format_.handle_response(
            response=response,
        )

        return response


class TablePageMixin(PaginationMixin, TemplateObjectMixin):
    page_keyword = "report_page"
    per_page_keyword = "report_per_page"
    page_template_keyword = "report_template"

    is_page_table = True
    have_template = True

    template_columns = None
    template_searchable_fields = None
    report_qs = None
    filters = None
    quicksearch = None
    used_filters = None
    ignore_search_values = [
        "unknown",
    ]
    ignore_search_keys = [
        "report_template",
    ]

    def get_template(self):
        page_template = self.request.GET.get(self.page_template_keyword)
        if page_template and (
            template := self.get_page_templates().filter(pk=page_template)
        ).exists():
            return template.first()
        
        return (
            self.get_page_templates().filter(is_page_default=True)
            or self.get_page_templates()
        ).first()

    def get_filters(self):
        initials = self.get_initials()

        self.template_filters = generate_filterset_from_model(
            self.report_model, self.get_form_classes()
        )(self.template_object.filters or {})
        self.filters = generate_filterset_from_model(
            self.report_model,
            self.get_form_classes(),
        )(initials)
        
        self.quicksearch = generate_quicksearch_filterset_from_model(
            self.report_model, list(self.template_searchable_fields.values())
        )(initials)

    def apply_user_path(self):
        LOGICAL_OPERATORS = ["()", "&", "|", "!="]
        paths = self.template_object.model_user_path or {}
        path_func = getattr(
            self.report_model,
            app_settings.MODEL_USER_PATH_FUNC_NAME,
            lambda request: {},
        )

        accessed_paths = {
            path_name: path_dict
            for path_name, path in paths.items()
            if (
                path_dict := {
                    path: call(path_func, request=self.request).get(path_name)
                }
            ).get(path)
        }

        if not len(accessed_paths):
            return

        accessed_path = (
            list(accessed_paths.keys())[0]
            if len(accessed_paths) == 1
            else list(filter(partial(ne, "__all__"), accessed_paths))[0]
        )
        if accessed_path == "__all__":
            return

        accessed_path, accessed_val = accessed_paths[accessed_path].popitem()
        if not any(map(lambda op: op in accessed_path, LOGICAL_OPERATORS)):
            self.report_qs = self.report_qs.filter(
                **{accessed_path: accessed_val}
            ).distinct()
            return

        self.report_qs = self.report_qs.filter(
            string_to_q(accessed_path, accessed_val)
        ).distinct()

    def _format_used_filter(self, col_name, val):
        formats = {
            **{k: "بله" for k in ["true", "True", True]},
            **{k: "خیر" for k in ["false", "False", False]},
        }

        if formatted_val := formats.get(val, False):
            return formatted_val

        if choices := get_choice_field_choices(self.report_model, col_name):
            return dict(choices).get(val, val)

        return str(val)

    def used_filter_format(self, col_name, val):
        if isinstance(val, list):
            return ", ".join(
                map(
                    lambda v: self._format_used_filter(col_name, v),
                    val,
                )
            )

        return self._format_used_filter(col_name, val) or val

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)

        obj = self.template_object
        if not obj:
            self.have_template = False
            return

        self.report_model = obj.model.model_class()
        self.template_columns = get_template_columns(obj, as_dict=False)
        self.template_searchable_fields = get_template_columns(obj, searchables=True)

        self.get_filters()
        self.report_qs = (
            self.template_filters.qs.distinct()
            if self.template_filters.get_filters()
            else self.report_model.objects.all()
        )
        self.apply_user_path()

        if (
            all(
                map(
                    methodcaller("get_filters"),
                    [self.filters, self.quicksearch, self.template_filters],
                )
            )
            and all(
                map(
                    attrgetter("data"),
                    [self.filters, self.quicksearch, self.template_filters],
                )
            )
            and all(
                map(
                    methodcaller("is_valid"),
                    [self.filters, self.quicksearch, self.template_filters],
                )
            )
        ):
            self.report_qs = self.template_filters.qs.distinct().filter(
                pk__in=Subquery(
                    (
                        self.quicksearch.qs.distinct() & self.filters.qs.distinct()
                    ).values("pk")
                )
            )

            self.report_qs = self.report_qs.distinct().order_by(
                *self.report_model._meta.ordering or ["pk"]
            )

            cleaned_data = (
                self.quicksearch.form.cleaned_data | self.filters.form.cleaned_data
            )
            self.used_filters = self.get_used_filters(
                {
                    get_column_verbose_name(
                        self.report_model, k
                    ): self.used_filter_format(k, v)
                    for k, v in cleaned_data.items()
                    if bool(v)
                }
            )

    def get_used_filters(self, cleaned_data):
        return _(" and ").join(
            [
                f'{k} = {",".join(map(str, v)) if isinstance(v, list) else v}'
                for k, v in cleaned_data.items()
                if str(k).lower() != "search"
            ]
        )

    def _prepare_initial(self, initial):
        if initial.lower() in ["true", "false"]:
            return initial.lower() == "true"

        if (initial.startswith("[") and initial.endswith("]")) or (
            not initial.startswith("0") and initial.isnumeric()
        ):
            return ast.literal_eval(initial)
        
        return initial

    def get_initial_value(self, initial, *, key=""):
        initial = str(initial)

        if key.endswith("__in"):
            return list(map(self._prepare_initial, self.request.GET.getlist(key)))

        return self._prepare_initial(initial)

    def get_initials(self):
        return {
            k: self.get_initial_value(v, key=k)
            for k, v in self.request.GET.dict().items()
            if str(v)
            and v.strip() not in self.ignore_search_values
            and k.strip() not in self.ignore_search_keys
        }

    def get_form_classes(self):
        if not self.template_object:
            return []
        return [generate_filterset_form(self.report_model)]

    def get_paginate_qs(self):
        return self.report_qs

    def get_context_data(self, **kwargs):
        if self.have_template:
            context = super().get_context_data(**kwargs)
        else:
            return super(TemplateObjectMixin, self).get_context_data(**kwargs)

        context["report"] = {
            "columns": self.template_columns,
            "columns_count": len(self.template_columns)
            + self.template_object.buttons.count()
            + sum(
                len(field.get_dynamic_obj().unpack_field())
                for field in self.template_columns.filter(
                    column_type=FieldTypes.dynamic
                ).only("pk")
            )
            + 1,
            "filters": self.filters,
            "buttons": self.template_object.buttons.all(),
            "searchable_fields": self.template_searchable_fields,
            "quicksearch": self.quicksearch,
            "used_filters": self.used_filters,
            "template": self.template_object,
            "templates": self.get_page_templates(),
            "initials": self.get_initials(),
            "pagination": self.pagination,
            "page_template_keyword": self.page_template_keyword,
            "is_page_table": self.is_page_table,
            "have_template": self.have_template,
            "export_formats": [
                {"name": format_.format_name, "slug": format_.format_slug}
                for format_ in export_format.formats.values()
            ],
            "page_title": getattr(
                self.template_object.page, "title", self.template_object.title
            ),
        }
        return context

    def get_page_templates(self):
        return Template.objects.filter(
            page__url_name=self.request.resolver_match.view_name
        ).order_by("-is_page_default")



import importlib

from phonenumber_field.modelfields import PhoneNumberField
from django.conf import settings
from django.db import models
from django_jalali.db import models as jmodels


def import_attribute(path):
    assert isinstance(path, str)
    pkg, attr = path.rsplit(".", 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret


def import_callable(path_or_callable):
    if not hasattr(path_or_callable, "__call__"):
        ret = import_attribute(path_or_callable)
    else:
        ret = path_or_callable
    return ret


def get_setting(name, dflt):
    getter = getattr(
        settings,
        "REPORT_SETTING_GETTER",
        lambda name, dflt: getattr(settings, name, dflt),
    )
    getter = import_callable(getter)
    return getter(name, dflt)


class AppSettings(object):
    def __init__(self, prefix="REPORT_"):
        self.prefix = prefix

    def _settings(self, name, dflt):
        return get_setting(self.prefix + name, dflt)

    @property
    def FILTERSET_CLASS(self):
        return import_callable(self._settings("FILTERSET_CLASS", "filterset.Filterset"))

    @property
    def FORMS(self):
        dflt = forms = {
            "CREATE_COLUMN": lambda form: form,
            "CREATE_TEMPLATE": lambda form: form,
        }
        if FORMS := self._settings("FORMS", False):
            forms = {k: FORMS.get(k, v) for k, v in dflt.items()}

        return {k: import_callable(v) for k, v in forms.items()}

    @property
    def BASE_VIEW(self):
        base_view = self._settings("BASE_VIEW", False)
        return import_callable(base_view) if base_view else None

    @property
    def DEFAULT_CELL_VALUE(self):
        return self._settings("DEFAULT_CELL_VALUE", "&mdash;")

    @property
    def EDITORS_GROUP_NAME(self):
        return self._settings("EDITORS_GROUP_NAME", "report_editors")

    @property
    def TIME_FORMATS(self):
        dflt = time_formats = {
            models.DateTimeField: "%H:%M %Y/%m/%d",
            models.DateField: "%Y/%m/%d",
            models.TimeField: "%H:%M:%S",
            jmodels.jDateField: "%H:%M:%S",
            jmodels.jDateTimeField: "%H:%M %Y/%m/%d",
        }

        if FORMATS := self._settings("TIME_FORMATS", False):
            assert isinstance(FORMATS, dict)
            time_formats = {k: FORMATS.get(k, v) for k, v in dflt.items()}

        return time_formats

    @property
    def DATA_TAGS(self):
        dflt = data_tags = {
            models.ImageField: lambda v: f'<img src="{v}" height=100>',
            models.FileField: lambda v: f'<a href="{v}" download="{(v_name:=v.split('/')[-1])}">{v_name}</a>',
            models.BooleanField: lambda v: f'<i class="material-icons-outlined">{v and 'check_circle' or 'cancel'}</i>',
            "default": lambda v: f"<span>{v}</span>",
            PhoneNumberField: lambda v: f"<span>{v.replace(' ', '-')}</span>",
        }

        if DATA_TAGS := self._settings("DATA_TAGS", False):
            assert isinstance(DATA_TAGS, dict)
            data_tags = {k: DATA_TAGS.get(k, v) for k, v in dflt.items()}

        return data_tags
    
    @property
    def MODEL_USER_PATH_FUNC_NAME(self):
        return self._settings("MODEL_USER_PATH_FUNC_NAME", "report_user_path")

    @property
    def VIEWS(self):
        dflt = views = {
            "TEMPLATE_LIST": "flex_report.views.template_list_view",
            "GENERAL_QS_EXPORT": "flex_report.views.general_qs_export_view",
            "TEMPLATE_CREATE_INIT": "flex_report.views.template_create_init_view",
            "TEMPLATE_CREATE_COMPLETE": "flex_report.views.template_create_complete_view",
            "TEMPLATE_DELETE": "flex_report.views.template_delete_view",
            "TEMPLATE_UPDATE": "flex_report.views.template_update_view",
            "TEMPLATE_CLONE": "flex_report.views.template_clone_view",
            "TEMPLATE_TOGGLE_DEFAULT": "flex_report.views.template_toggle_default_view",
            "COLUMN_LIST": "flex_report.views.column_list_view",
            "COLUMN_CREATE": "flex_report.views.column_create_view",
            "COLUMN_UPDATE": "flex_report.views.column_update_view",
            "COLUMN_DELETE": "flex_report.views.column_delete_view",
            "REPORT": "flex_report.views.report_view",
            "REPORT_EXPORT": "flex_report.views.report_export_view",
        }
        if VIEWS := self._settings("VIEWS", False):
            assert isinstance(VIEWS, dict)
            views = {k: VIEWS.get(k, v) for k, v in dflt.items()}

        return {k: import_callable(v) for k, v in views.items()}

    @property
    def REALTIME_QUICKSEARCH(self):
        return self._settings("REALTIME_QUICKSEARCH", True)


app_settings = AppSettings()
